#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <syslog.h>
#include <wiringPi.h>
#include <string.h>
#include <pthread.h>

#define GPIO_ALARM 		 26
#define GPIO_FI_END 	 13
#define GPIO_FI_YAM 	 19
#define GPIO_CAMLI_TOG 	 5
#define GPIO_EXLI_TOG 	 6
#define GPIO_RLY_1 		 12
#define GPIO_RLY_2_EXTLI 16
#define GPIO_RLY_3_CAMLI 20
#define GPIO_RLY_4_PRN 	 21

#define ISRDELAY_MS  1000

typedef int bool;

//deamon stuff PID handling
#define PROGNAME "gpioservice"
#define PIDFILE "/var/run/" PROGNAME ".pid"
#define CPF_CLOEXEC FD_CLOEXEC
#define BUF_SIZE 512

/* Datenpuffer fuer die GPIO-Funktionen */
#define MAXBUFFER 100 
#define TC_LOCK 1

typedef struct THREAD_PARAMS{
    volatile bool  runThread;
    pthread_t       pthID;
    pthread_mutex_t cond_mtx;
    pthread_cond_t  wait_cond;
	int				pin;
	char			name[32];
}THREAD_PARAMS;

THREAD_PARAMS exLiTogThParams;
THREAD_PARAMS camLiTogThParams;
struct timespec tsLastExLi;
struct timespec tsLastCamLi;
bool runLoop;

//suporters
int delayMs(unsigned long millis) {
	struct timespec ts;
	int err;

	ts.tv_sec = millis / 1000;
	ts.tv_nsec = (millis % 1000) * 1000000L;
	err = nanosleep(&ts, (struct timespec *)NULL);
	return (err);
}

ulong getDeltaMs( struct timespec tsStart )
{
    long            ms; // Milliseconds
    long            nsMs; // NanoSeconds
	struct timespec tsNow;
    clock_gettime(CLOCK_REALTIME, &tsNow);
	
	ms = (tsNow.tv_sec - tsStart.tv_sec) * 1000;
	nsMs = round( (tsNow.tv_nsec - tsStart.tv_nsec) / 1000000 );
	//printf( "getDeltaMs ms = %ld, nsMs = %ld \n", ms, nsMs);
	ms += nsMs;
	return (ms < 0 ? 0 : ms);
}

bool gpio_aktivate(int pin) {
	char buffer[MAXBUFFER];    /* Output Buffer   */
	ssize_t bytes;             /* Datensatzlaenge */
	int fd;                    /* Filedescriptor  */
	int res;                   /* Ergebnis von write */

	snprintf(buffer, MAXBUFFER, "/sys/class/gpio/gpio%d/value", pin);
	fd = open(buffer, O_WRONLY);
	if (fd > 0) {
		close(fd);
		//printf ("pin %d already exported! \n", pin );
		syslog (LOG_ERR, "pin %d already exported!", pin );				
		return(TRUE);
    }
	
	fd = open("/sys/class/gpio/export", O_WRONLY);
	if (fd < 0){
		//printf ("Kann nicht auf export schreiben: %d \n", errno );
		syslog (LOG_ERR, "Kann nicht auf export schreiben: %d", errno );						
		return(FALSE);
    }
	
	bytes = snprintf(buffer, MAXBUFFER, "%d", pin);
	res = write(fd, buffer, bytes);
	close(fd);
		
	if (res < 0){
		//printf ("Kann Pin nicht aktivieren (write): %d \n", errno );
		syslog (LOG_ERR, "Kann Pin nicht aktivieren (write): %d", errno );								
		return(FALSE);
    }

	delay(100);
	return(TRUE);
}

bool gpio_write(int pin, int value) {
	char path[MAXBUFFER];      /* Buffer fuer Pfad   */
	int fd;                    /* Filedescriptor     */
	int res;                   /* Ergebnis von write */

	snprintf(path, MAXBUFFER, "/sys/class/gpio/gpio%d/value", pin);
	fd = open(path, O_WRONLY);
	if (fd < 0) {
		//printf ("Kann auf GPIO nicht schreiben (open): %d \n", errno );
		syslog (LOG_ERR, "Kann auf GPIO nicht schreiben (open): %d", errno );		
		return(FALSE);
    }
	
	switch (value) {
		case LOW : res = write(fd,"0",1); break;
		case HIGH: res = write(fd,"1",1); break;
    }
	close(fd);
	
	if (res < 0) {
		//printf ("Kann auf GPIO nicht schreiben (write): %d \n", errno );
		syslog (LOG_ERR, "Kann auf GPIO nicht schreiben (write): %d", errno );				
		return(FALSE);
    }
  
  return(TRUE);
 }
 
 int gpio_read(int pin)
 {
 	char path[MAXBUFFER];      /* Buffer fuer Pfad   */
	int fd;                    /* Filedescriptor     */
	char buffer[10];           /* Ergebnis von read */

	snprintf(path, MAXBUFFER, "/sys/class/gpio/gpio%d/value", pin);
	fd = open(path, O_RDONLY);
	if (fd < 0) {
		//printf ("Kann GPIO nicht lesen (open): %d \n", errno );
		syslog (LOG_ERR, "Kann GPIO nicht lesen (open): %d", errno );						
		return(-1);
    }
    
	if( read(fd, buffer, sizeof(buffer)-1) > 0){
		close(fd);
		return atoi(buffer);
	}
    close(fd);
	//printf ("Kann GPIO nicht lesen (read): %d \n", errno );
	syslog (LOG_ERR, "Kann GPIO nicht lesen (read): %d", errno );						
	return(-1);
 }
 
bool gpio_setMode(int pin, int dir, int pr) {
	char path[MAXBUFFER];      /* Buffer fuer Pfad   */
	int fd;                    /* Filedescriptor     */
	int res;                   /* Ergebnis von write */

	snprintf(path, MAXBUFFER, "/sys/class/gpio/gpio%d/direction", pin);
	fd = open(path, O_WRONLY);
	if (fd < 0) {
		//printf ("Kann Datenrichtung nicht setzen (open): %d \n", errno );
		syslog (LOG_ERR, "Kann Datenrichtung nicht setzen (open): %d", errno );								
		return(FALSE);
    }
  
	switch (dir) {
		case INPUT : res = write(fd,"in",2); break;
		case OUTPUT: res = write(fd,"out",3); break;
    }
	close(fd);
	
	if (res < 0) {
		//printf ("Kann Datenrichtung nicht setzen (write): %d \n", errno );
		syslog (LOG_ERR, "Kann Datenrichtung nicht setzen (write): %d", errno );								
		return(FALSE);
	}
  
    switch (dir) {
		//set pull R to IN pins or write init val to OUT pins
	    case INPUT : pullUpDnControl (pin, pr); break;
		case OUTPUT: gpio_write(pin, pr ); break;
	}
  
  return(TRUE);
 }

int initPins( void )
{
	//relay board
	if( ! ( gpio_aktivate( GPIO_RLY_1 ) && gpio_setMode( GPIO_RLY_1, OUTPUT, HIGH) )) {
		//printf ("error at setup GPIO_RLY_1 \n" );
		syslog (LOG_ERR, "error at setup GPIO_RLY_1" );
	} 	

	if( ! ( gpio_aktivate( GPIO_RLY_2_EXTLI ) && gpio_setMode( GPIO_RLY_2_EXTLI, OUTPUT, HIGH ) )) {
		//printf ("error at setup GPIO_RLY_2_EXTLI \n" );
		syslog (LOG_ERR, "error at setup GPIO_RLY_2_EXTLI" );
	} 	

	if( ! ( gpio_aktivate( GPIO_RLY_3_CAMLI ) && gpio_setMode( GPIO_RLY_3_CAMLI, OUTPUT, HIGH ) )) {
		//printf ("error at setup GPIO_RLY_3_CAMLI \n" );
		syslog (LOG_ERR, "error at setup GPIO_RLY_3_CAMLI" );
	} 	

	if( ! ( gpio_aktivate( GPIO_RLY_4_PRN ) && gpio_setMode( GPIO_RLY_4_PRN, OUTPUT, HIGH ) )) {
		printf ("error at setup GPIO_RLY_4_PRN \n" );
		syslog (LOG_ERR, "error at setup GPIO_RLY_4_PRN" );
	} 	

	//alarm
	if( ! ( gpio_aktivate( GPIO_ALARM ) && gpio_setMode( GPIO_ALARM, INPUT, PUD_OFF ) )) {
		//printf ("error at setup GPIO_ALARM \n" );
		syslog (LOG_ERR, "error at setup GPIO_ALARM" );
	} 	
	
	//filament
	if( ! ( gpio_aktivate( GPIO_FI_END ) && gpio_setMode( GPIO_FI_END, INPUT, PUD_OFF ) )) {
		//printf ("error at setup GPIO_FI_END \n" );
		syslog (LOG_ERR, "error at setup GPIO_FI_END" );
	} 	
	
	if( ! ( gpio_aktivate( GPIO_FI_YAM ) && gpio_setMode( GPIO_FI_YAM, INPUT, PUD_OFF ) )) {
		//printf ("error at setup GPIO_FI_YAM \n" );
		syslog (LOG_ERR, "error at setup GPIO_FI_YAM" );
	} 	
	
	// toggle relay
	if( ! ( gpio_aktivate( GPIO_CAMLI_TOG ) && gpio_setMode( GPIO_CAMLI_TOG, INPUT, PUD_OFF ) )) {
		//printf ("error at setup GPIO_PRN_TOG \n" );
		syslog (LOG_ERR, "error at setup GPIO_PRN_TOG" );
	} 	
	
	if( ! ( gpio_aktivate( GPIO_EXLI_TOG ) && gpio_setMode( GPIO_EXLI_TOG, INPUT, PUD_OFF ) )) {
		//printf ("error at setup GPIO_EXLI_TOG \n" );
		syslog (LOG_ERR, "error at setup GPIO_EXLI_TOG" );
	} 	
	
	return TRUE;
}

//isr for camera light toggle
void gpioInterruptCamLight(void) {
    //printf( "ISR camera light toggle called \n" );
    if( ISRDELAY_MS > getDeltaMs( tsLastCamLi )) {
		return;
	}
	clock_gettime(CLOCK_REALTIME, &tsLastCamLi);
    if( (camLiTogThParams.pthID != 0) && camLiTogThParams.runThread ){
         pthread_mutex_lock(&(camLiTogThParams.cond_mtx));
         pthread_cond_signal(&(camLiTogThParams.wait_cond));
         pthread_mutex_unlock(&(camLiTogThParams.cond_mtx));
    } else {
		//printf ("ISR camera light toggle failed! Worker thread not active \n" );
		syslog (LOG_ERR, "ISR camera light toggle failed! Worker thread not active" );		
	}   
}

//isr for extruder light toggle
void gpioInterruptExLight(void) {
    //printf( "ISR extruder light toggle called \n" );
    if( ISRDELAY_MS > getDeltaMs( tsLastExLi )) {
		return;
	}
	clock_gettime(CLOCK_REALTIME, &tsLastExLi);
	if( (exLiTogThParams.pthID != 0) && exLiTogThParams.runThread ){
         pthread_mutex_lock(&(exLiTogThParams.cond_mtx));
         pthread_cond_signal(&(exLiTogThParams.wait_cond));
         pthread_mutex_unlock(&(exLiTogThParams.cond_mtx));
    } else {
		//printf ("ISR extruder light toggle failed! Worker thread not active \n" );
		syslog (LOG_ERR, "ISR extruder light toggle failed! Worker thread not active" );		
	}
}

void * ToggleWorkerThread(void * initData) {
	THREAD_PARAMS * thParams = (THREAD_PARAMS *)initData;
    while(thParams->runThread){
	    delayMs(100);
        pthread_mutex_lock(&(thParams->cond_mtx));
        pthread_cond_wait( &(thParams->wait_cond), &(thParams->cond_mtx));
        pthread_mutex_unlock(&(thParams->cond_mtx));

        if( thParams->runThread ){
			//printf ("%sThread do a run", thParams->name );
			int gpioval = gpio_read(thParams->pin);
			if(gpioval == 1){
				gpioval = 0;
			} else if( gpioval == 0 ) {
				gpioval = 1;
			} else {
				//printf ("%s worker thread GPIO read error: %d \n", thParams->name, errno );
				syslog (LOG_ERR, "%s worker thread GPIO read error: %d", thParams->name, errno );
				continue;
			}
			delayMs(100);
			if( ! gpio_write( thParams->pin, gpioval) ) {
				//printf ("%s worker thread GPIO write error: %d \n", thParams->name, errno );
				syslog (LOG_ERR, "%s worker thread GPIO write error: %d", thParams->name, errno );			
			}
			continue;
        }
    }
    thParams->pthID = 0L;
    pthread_exit( NULL );
}

int gpioserv_main( void )
{
	wiringPiSetupGpio();
	
	if( ! initPins() ) {
		//printf ("initPins() failed: %d \n", errno );
		syslog (LOG_ERR, "initPins() failed: %d", errno );		
	}
		
	camLiTogThParams.runThread = TRUE;
    camLiTogThParams.pthID = 0L;
    camLiTogThParams.cond_mtx = (pthread_mutex_t)PTHREAD_MUTEX_INITIALIZER;
    camLiTogThParams.wait_cond = (pthread_cond_t)PTHREAD_COND_INITIALIZER;
	camLiTogThParams.pin = GPIO_RLY_3_CAMLI;
	sprintf( camLiTogThParams.name, "CamLight" );
	clock_gettime(CLOCK_REALTIME, &tsLastCamLi);	

	if( 0 != pthread_create(&(camLiTogThParams.pthID), NULL, ToggleWorkerThread, (void *)(&camLiTogThParams)) ){
		camLiTogThParams.pthID = 0L;
		//printf ("create  camera light toggle failed: %d \n", errno );
		syslog (LOG_ERR, "create  camera light toggle failed: %d", errno );			
	} else {
		if( wiringPiISR ( GPIO_CAMLI_TOG, INT_EDGE_FALLING, &gpioInterruptCamLight) < 0 ){
			//printf ("Unable to setup camera light toggle ISR: %d \n", errno );
			syslog (LOG_ERR, "Unable to setup camera light toggle ISR: %d", errno );	
		}	
	}

	exLiTogThParams.runThread = TRUE;
    exLiTogThParams.pthID = 0L;
    exLiTogThParams.cond_mtx = (pthread_mutex_t)PTHREAD_MUTEX_INITIALIZER;
    exLiTogThParams.wait_cond = (pthread_cond_t)PTHREAD_COND_INITIALIZER;
	exLiTogThParams.pin = GPIO_RLY_2_EXTLI;
	sprintf( camLiTogThParams.name, "ExLight" );
	clock_gettime(CLOCK_REALTIME, &tsLastExLi);
	
	if( 0 != pthread_create(&(exLiTogThParams.pthID), NULL, ToggleWorkerThread, (void *)(&exLiTogThParams)) ){
		exLiTogThParams.pthID = 0L;
		//printf ("create  power toggle failed: %d \n", errno );
		syslog (LOG_ERR, "create  power toggle failed: %d", errno );			
	} else {
		if( wiringPiISR ( GPIO_EXLI_TOG, INT_EDGE_FALLING, &gpioInterruptExLight) < 0 ){
			//printf ("Unable to setup extr. light toggle ISR: %d \n", errno );
			syslog (LOG_ERR, "Unable to extr. light toggle toggle ISR: %d", errno );	
		}	
	}
		
	//printf( "GPIO service started\n" );
	syslog (LOG_INFO, "GPIO service started" );
	
	runLoop = TRUE;
		
	//int loopcout = 0;
	while( runLoop ) {
	    //sleep( 10 );
		//printf( "GPIO service still running\n" );
		pause();
	}
	
	if( (camLiTogThParams.pthID != 0) && camLiTogThParams.runThread ){
		camLiTogThParams.runThread = FALSE;
		gpioInterruptCamLight();
	}
		
	if( (exLiTogThParams.pthID != 0) && exLiTogThParams.runThread ){
		exLiTogThParams.runThread = FALSE;
		gpioInterruptExLight();
	}
		
	//printf( "GPIO service  finished\n" );
	syslog (LOG_INFO, "GPIO service finished" );
	return 0;
}

//demon
int createPidFile(const char *progName, const char *pidFile, int flags)
{
    int fd;
    char buf[BUF_SIZE];

    fd = open(pidFile, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
    if (fd == -1) {
        syslog (LOG_ERR, "Could not open PID file %s", pidFile );
		return 1;
	}

    if (flags & CPF_CLOEXEC) {
        flags = fcntl(fd, F_GETFD);                     // Fetch flags
        if (flags == -1){
            syslog (LOG_ERR,"Could not get flags for PID file %s", pidFile);
			return 1;
		}

        flags |= FD_CLOEXEC;                            // Turn on FD_CLOEXEC 

        if (fcntl(fd, F_SETFD, flags) == -1){            // Update flags 
            syslog (LOG_ERR, "Could not set flags for PID file %s", pidFile);
			return 1;
		}
    }

	if( fcntl(fd, F_WRLCK) == -1 ) {
        syslog (LOG_ERR, "PID file '%s' is locked; probably '%s' is already running", pidFile, progName);
		return 1;
    }

    if (ftruncate(fd, 0) == -1) {
        syslog (LOG_ERR,"Could not truncate PID file '%s'", pidFile);
	}

    snprintf(buf, BUF_SIZE, "%ld\n", (long) getpid());
    if (write(fd, buf, strlen(buf)) != strlen(buf)) {
        syslog (LOG_ERR, "Writing to PID file '%s'", pidFile);
		return 2;
	}
	
    return 0;
}

int main(void) 
{		
	setlogmask(LOG_UPTO (LOG_INFO));
	openlog(PROGNAME, LOG_CONS | LOG_PID | LOG_NDELAY, LOG_DAEMON);
	syslog (LOG_INFO, PROGNAME " starting" );
	
	pid_t pid, sid;
	
	/* Fork off the parent process */
	pid = fork();
	if (pid < 0) {
			exit(EXIT_FAILURE);
	}

	if (pid > 0) {
			exit(EXIT_SUCCESS);
	}

	umask(0);
	sid = setsid();
	if (sid < 0) {
			exit(EXIT_FAILURE);
	}
	
	if ((chdir("/")) < 0) {
			exit(EXIT_FAILURE);
	}
	
	if( 0 != createPidFile(PROGNAME, PIDFILE, CPF_CLOEXEC)) {
		syslog (LOG_WARNING, "error on creating PID file for " PROGNAME );
	}
	
	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	close(STDERR_FILENO);
	
	/* Daemon-specific working stuff */
	int iRes = gpioserv_main();

	syslog (LOG_INFO, PROGNAME " finished" );
	closelog();
    exit(iRes);
}